/*
  cmdSource.c
  Jeff McElroy
  jmcelroy11@sio.midco.net
*/

/*
    ypAnything, an NIS gateway to non-NIS sources.
    Copyright (C) 2004  Jeff McElroy <jmcelroy11@sio.midco.net>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                                                                 
*/

#include <pthread.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <assert.h>
#include <stdio.h>
#include <errno.h>

#include "heap.h"
#include "mapping.h"
#include "data.h"
#include "log.h"
#include "cmdSource.h"

#define MALLOC_SIZE (2500 * 1024)
#define FREAD_SIZE  (64 * 1024)

typedef struct CMDSOURCE_DATA
{
  pthread_t m_pthread;		/* this thread */
  int m_refreshRate;		/* number of seconds to wait between refreshes */
  char *m_sourceProgram;	/* program to run to refresh the data */
  char *m_deliminators;		/* deliminator characters */
  char *m_data;			/* in memory copy of the map */
} CMDSOURCE_DATA;


/******************************************************************************
  cmdsource index 
******************************************************************************/
typedef struct CMDSOURCEINDEX_DATA
{
  int m_keyIndex;
  MAPPING *m_mapping;
} CMDSOURCEINDEX_DATA;


DATUM
cmdsourceIndexGetFirst (struct DOMAIN *domain, struct MAP *map,
			struct INDEX *index, DATUM * d)
{
  struct CMDSOURCEINDEX_DATA *data = index->m_data;
  DATUM result;

  if (mappingGet
      (data->m_mapping, &d->m_ptr, &d->m_size, &result.m_ptr,
       &result.m_size) == 1) {
    result.m_ptr = NULL;
    result.m_size = 0;
  }
  return result;
}

DATUM
cmdsourceIndexGetNext (struct DOMAIN * domain, struct MAP * map,
		       struct INDEX * index, DATUM * d)
{
  CMDSOURCEINDEX_DATA *data = index->m_data;
  DATUM result;

  if (mappingGetNext
      (data->m_mapping, &(d->m_ptr), &(d->m_size), &result.m_ptr,
       &result.m_size) == 1) {
    result.m_ptr = NULL;
    result.m_size = 0;
  }

  return result;
}

static void cmdsourceIndexDataClose (CMDSOURCEINDEX_DATA * data)
{
  mappingClose (data->m_mapping);
  heapFree (data, MEMTYPE_CMDSOURCE_INDEXDATA);
}

static void
cmdsourceIndexOpen (struct DOMAIN *domain, struct MAP *map, char *indexName,
		    int keyIndex)
{
  INDEX *index;
  struct CMDSOURCEINDEX_DATA *data;

  index = dataIndexAdd (map, indexName);

  index->m_getFirstFunc = cmdsourceIndexGetFirst;
  index->m_getNextFunc = cmdsourceIndexGetNext;

  data =
    (struct CMDSOURCEINDEX_DATA *) heapAlloc (sizeof (CMDSOURCEINDEX_DATA),
					      MEMTYPE_CMDSOURCE_INDEXDATA);
  assert (data != NULL);

  data->m_mapping = mappingOpen ();
  data->m_keyIndex = keyIndex;

  index->m_data = data;
}

static CMDSOURCEINDEX_DATA *cmdsourceIndexRefresh (struct DOMAIN *domain,
						   MAP * map, INDEX * index,
						   char *cmdsourceText)
{
  CMDSOURCEINDEX_DATA *indexData = index->m_data;
  CMDSOURCE_DATA *mapData = map->m_data;
  char *deliminator;
  char *record;
  int recordLength;
  char *key;
  int keyLength;
  int field;

  mappingClose (indexData->m_mapping);
  indexData->m_mapping = mappingOpen ();
  record = cmdsourceText;
  deliminator = mapData->m_deliminators;

   /*REDO*/ while (*record != '\0') {
    recordLength = strcspn (record, "\n");

    /* get the key */
    key = record;
    for (field = 1; field < indexData->m_keyIndex; field++) {
      key += strcspn (key, deliminator);
      if (*key != '\0') {
	key++;
      }
    }
    keyLength = strcspn (key, mapData->m_deliminators);
    mappingSet (indexData->m_mapping, key, keyLength, record, recordLength);
    record += recordLength;

    if (*record == '\n') {
      record++;
    }

  }
  return indexData;
}

/******************************************************************************
cmdsource map
******************************************************************************/

static void cmdsourceRefresh (struct DOMAIN *domain, struct MAP *map)
{
  FILE *file;
  CMDSOURCE_DATA *mapData = map->m_data;

  INDEX *index;

  char *oldData;
  char *newData;

  struct stat statScript;
  int i;
  int totalRead;
  int totalMemory;
  int rc;

  log (LOG_STATUS, "cmdsource refresh started for %s.%s", domain->m_name,
       map->m_name);

#if defined (STAT_THE_SCRIPT)
  if (stat (mapData->m_sourceProgram, &statScript) < 0) {
    log (LOG_ERROR, "error, stat(%s): '%s'", mapData->m_sourceProgram,
	 strerror (errno));
    return;
  }
  else if (statScript.st_uid != getuid ()) {
    log (LOG_ERROR,
	 "error, The script '%s' must be owned by the same user that ypanything runs under, please 'chown' it.",
	 mapData->m_sourceProgram);
    return;
  }
  else if ((statScript.st_mode & S_IXUSR) != S_IXUSR) {
    log (LOG_ERROR,
	 "error, %s must be executible by it's owner, please 'chmod' it",
	 mapData->m_sourceProgram);
    return;
  }
#endif

  file = popen (mapData->m_sourceProgram, "r");
  if (file == NULL) {
    log (LOG_ERROR, "error, popen(%s): '%s'", mapData->m_sourceProgram,
	 strerror (errno));
    return;
  }

  oldData = mapData->m_data;
  totalMemory = MALLOC_SIZE;
  newData = (char *) heapAlloc (MALLOC_SIZE, MEMTYPE_CMDSOURCE_MAPTEXT);
  assert (newData != NULL);
  totalRead = 0;
  do {
    if (totalRead + FREAD_SIZE >= totalMemory) {
      totalMemory += MALLOC_SIZE;
      newData = heapRealloc (newData, totalMemory, MEMTYPE_CMDSOURCE_MAPTEXT);
    }
    i = fread (newData + totalRead, sizeof (char), FREAD_SIZE, file);
    totalRead += i;
  }
  while (i > 0);

  *(newData + totalRead) = '\0';
  rc = pclose (file);

  if (rc != 0) {
    log (LOG_ERROR,
	 "error, the script '%s' returned with an error code of %d",
	 mapData->m_sourceProgram, rc);
    heapFree (newData, MEMTYPE_CMDSOURCE_MAPTEXT);
    return;
  }

  /* this part needs to be atomic and fast enough that it doesnt matter */
  {
    dataMapLock (map);
    log (LOG_DEBUG, "cmdsource critical area started");

    for (index = map->m_index; index != NULL; index = index->m_next) {
      log (LOG_DEBUG, "refreshing index %s", index->m_name);
      cmdsourceIndexRefresh (domain, map, index, newData);
    }

    mapData->m_data = newData;
    if (oldData != NULL) {
      heapFree (oldData, MEMTYPE_CMDSOURCE_MAPTEXT);
    }
    dataMapUnlock (map);
    log (LOG_DEBUG, "cmdsource critical area completed");
  }
  log (LOG_STATUS, "cmdsource refresh completed for %s.%s", domain->m_name,
       map->m_name);
}

typedef struct REFRESH_THREAD_DATA
{
  DOMAIN *m_domain;
  MAP *m_map;
  int m_refreshRate;
} REFRESH_THREAD_DATA;

static void *cmdsourceRefreshLoop (void *arg)
{
  REFRESH_THREAD_DATA *threadData = arg;

  while (1) {
    cmdsourceRefresh (threadData->m_domain, threadData->m_map);
    sleep (threadData->m_refreshRate);
  }
  heapFree (threadData, MEMTYPE_CMDSOURCE_THREADDATA);

  return NULL;

}

static void
cmdsourceOpen (struct DOMAIN *domain, struct MAP *map, char *sourceProgram,
	       char *deliminators, int indexCount, char *indexNames[],
	       int indexFields[])
{
  struct CMDSOURCE_DATA *data;
  int i;
  data =
    (struct CMDSOURCE_DATA *) heapAlloc (sizeof (CMDSOURCE_DATA),
					 MEMTYPE_CMDSOURCE_MAPDATA);
  assert (data != NULL);
  data->m_data = NULL;
  data->m_sourceProgram =
    heapStrdup (sourceProgram, MEMTYPE_CMDSOURCE_SCRIPT);
  data->m_deliminators =
    heapStrdup (deliminators, MEMTYPE_CMDSOURCE_DELIMINATOR);

  for (i = 0; i < indexCount; i++) {
    cmdsourceIndexOpen (domain, map, indexNames[i], indexFields[i]);
  }
  map->m_data = (void *) data;
}

static void cmdsourceClose (struct DOMAIN *domain, struct MAP *map)
{
  CMDSOURCE_DATA *data = map->m_data;
  if (data->m_data != NULL) {
    heapFree (data->m_sourceProgram, MEMTYPE_CMDSOURCE_SCRIPT);
    heapFree (data->m_deliminators, MEMTYPE_CMDSOURCE_DELIMINATOR);
    heapFree (data->m_data, MEMTYPE_CMDSOURCE_MAPTEXT);
  }
  heapFree (data, MEMTYPE_CMDSOURCE_MAPDATA);
}

MAP *cmdSourceMapAdd (DOMAIN * domain, char *name, int refreshRate,
		      char *sourceProgram, char *deliminators, int indexCount,
		      char *indexNames[], int indexFields[])
{
  MAP *map;
  REFRESH_THREAD_DATA *threadData;
  CMDSOURCE_DATA *cmdsourceData;

  map = dataMapAdd (domain, name);

  map->m_closeFunc = cmdsourceClose;

  cmdsourceOpen (domain, map, sourceProgram, deliminators, indexCount,
		 indexNames, indexFields);

  if (refreshRate == 0) {
    cmdsourceRefresh (domain, map);
  }
  else {
    threadData =
      (REFRESH_THREAD_DATA *) heapAlloc (sizeof (REFRESH_THREAD_DATA),
					 MEMTYPE_CMDSOURCE_THREADDATA);
    assert (threadData != NULL);

    threadData->m_domain = domain;
    threadData->m_map = map;
    threadData->m_refreshRate = refreshRate;
    cmdsourceData = (CMDSOURCE_DATA *) map->m_data;
    pthread_create (&cmdsourceData->m_pthread, NULL, cmdsourceRefreshLoop,
		    threadData);
  }
  return map;

}
