/*
  data.c
  Jeff McElroy
  jmcelroy11@sio.midco.net
*/

/*
    ypanything, an NIS gateway to non-NIS sources
    Copyright (C) 2004  Jeff McElroy <jmcelroy11@sio.midco.net>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                                                                 
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>

#include "heap.h"
#include "log.h"
#include "data.h"

#if !defined(MIN)
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

DATA *dataOpen (void)
{
  DATA *data;

  data = heapAlloc (sizeof (DATA), MEMTYPE_DATA);
  assert (data != NULL);
  data->m_domain = NULL;
  return data;
}

void dataClose (DATA * data)
{
  while (data->m_domain != NULL) {
    dataDomainDelete (data, data->m_domain);
  }
  heapFree (data, MEMTYPE_DATA);
}

DOMAIN *dataDomainAdd (DATA * data, char *name)
{
  DOMAIN *domain;

  domain = heapAlloc (sizeof (DOMAIN), MEMTYPE_DOMAIN);
  assert (domain != NULL);

  domain->m_name = heapStrdup (name, MEMTYPE_DOMAINNAME);
  assert (domain->m_name != NULL);

  domain->m_map = NULL;

  domain->m_next = data->m_domain;
  data->m_domain = domain;

  return domain;
}

void dataDomainDelete (DATA * data, DOMAIN * domain)
{
  DOMAIN **pp;

  pp = &data->m_domain;
  while (*pp != NULL && *pp != domain) {
    pp = &(*pp)->m_next;
  }

  if (*pp != NULL) {
    *pp = domain->m_next;

    while (domain->m_map != NULL) {
      dataMapDelete (domain, domain->m_map);
    }
    heapFree (domain->m_name, MEMTYPE_DOMAINNAME);
    heapFree (domain, MEMTYPE_DOMAIN);
  }
}

DOMAIN *dataDomainFind (DATA * data, char *name)
{
  DOMAIN *domain;
  domain = data->m_domain;
  while (domain != NULL && strcmp (domain->m_name, name) != 0) {
    domain = domain->m_next;
  }
  return domain;
}

void dataMapLock (MAP * map)
{
  log (LOG_DEBUG, "dataMapLock(%s)", map->m_name);
  pthread_mutex_lock (&map->m_mutext);
}

void dataMapUnlock (MAP * map)
{
  log (LOG_DEBUG, "dataMapUnLock(%s)", map->m_name);
  pthread_mutex_unlock (&map->m_mutext);
}

MAP *dataMapAdd (DOMAIN * domain, char *name)
{
  MAP *map;

  map = heapAlloc (sizeof (MAP), MEMTYPE_MAP);
  assert (map != NULL);

  map->m_name = heapStrdup (name, MEMTYPE_MAPNAME);
  assert (name != NULL);

  pthread_mutex_init (&map->m_mutext, NULL);
  map->m_index = NULL;
  map->m_openFunc = NULL;
  map->m_closeFunc = NULL;

  map->m_next = domain->m_map;
  domain->m_map = map;

  return map;
}

void dataMapDelete (DOMAIN * domain, MAP * map)
{
  MAP **pp;

  pp = &domain->m_map;
  while (*pp != NULL && *pp != map) {
    pp = &((*pp)->m_next);
  }

  if (*pp != NULL) {
    if (map->m_closeFunc != NULL) {
      (*map->m_closeFunc) (domain, map);
    }
    *pp = map->m_next;
    while (map->m_index != NULL) {
      dataIndexDelete (map, map->m_index);
    }

    heapFree (map->m_name, MEMTYPE_MAPNAME);
    heapFree (map, MEMTYPE_MAP);
  }
}

MAP *dataMapFind (DOMAIN * domain, char *name)
{
  MAP *map;
  map = domain->m_map;
  while (map != NULL && strcmp (map->m_name, name) != 0) {
    map = map->m_next;
  }
  return map;
}

INDEX *dataIndexAdd (MAP * map, char *name)
{
  INDEX *index;

  index = heapAlloc (sizeof (INDEX), MEMTYPE_INDEX);
  assert (index != NULL);

  index->m_name = heapStrdup (name, MEMTYPE_INDEXNAME);
  assert (name != NULL);

  index->m_next = map->m_index;
  map->m_index = index;

  return index;
}

void dataIndexDelete (MAP * map, INDEX * index)
{
  INDEX **pp;

  pp = &map->m_index;
  while (*pp != NULL && *pp != index) {
    pp = &((*pp)->m_next);
  }

  if (*pp != NULL) {
    *pp = index->m_next;
    heapFree (index->m_name, MEMTYPE_INDEXNAME);
    heapFree (index, MEMTYPE_INDEX);
  }
}

INDEX *dataIndexFind (MAP * map, char *name)
{
  INDEX *index;

  index = map->m_index;
  while (index != NULL && strcmp (index->m_name, name) != 0) {
    index = index->m_next;
  }

  return index;
}

void dataDump (DATA * data)
{
  DOMAIN *domain;
  MAP *map;
  INDEX *index;

  for (domain = data->m_domain; domain != NULL; domain = domain->m_next) {
    printf ("domain '%s' {\n", domain->m_name);
    for (map = domain->m_map; map != NULL; map = map->m_next) {
      printf ("\tmap '%s' {\n", map->m_name);
      for (index = map->m_index; index != NULL; index = index->m_next) {
	printf ("\t\tindex '%s' {\n", index->m_name);
	printf ("\t\t}\n");
      }
      printf ("\t}\n");
    }
    printf ("}\n");
  }
}

MAP *dataMapFromName (DOMAIN * domain, char *name)
{
  MAP *map;
  char *pc;
  char mapName[128];
  int len;

  map = NULL;
  pc = strchr (name, '.');
  if (pc != NULL) {
    len = MIN (pc - name, sizeof (mapName) - 1);
    memcpy (mapName, name, len);
    mapName[len] = '\0';
    map = dataMapFind (domain, mapName);
  }

  return map;
}

INDEX *dataIndexFromName (MAP * map, char *name)
{
  INDEX *index;
  char *pc;
  char indexName[128];

  index = NULL;
  pc = strchr (name, '.');
  if (pc != NULL) {
    strncpy (indexName, pc + 1, sizeof (indexName));
    index = dataIndexFind (map, indexName);
  }

  return index;
}

int
dataDomainMapAndIndexFromName (DATA * data, char *domainName,
			       char *mapIndexName, DOMAIN ** domainReturn,
			       MAP ** mapReturn, INDEX ** indexReturn)
{
  DOMAIN *domain = NULL;
  MAP *map = NULL;
  INDEX *index = NULL;
  int rc = 1;

  domain = dataDomainFind (data, domainName);
  if (domain == NULL) {
    log (LOG_DEBUG,
	 "dataDomainMapAndIndexFromName(%s,%s) {domain not found}",
	 domainName, mapIndexName);
  }
  else {
    map = dataMapFromName (domain, mapIndexName);
    if (map == NULL) {
      log (LOG_DEBUG,
	   "dataDomainMapAndIndexFromName(%s,%s) {map not found}",
	   domainName, mapIndexName);
    }
    else {
      index = dataIndexFromName (map, mapIndexName);
      if (index == NULL) {
	log (LOG_DEBUG,
	     "dataDomainMapAndIndexFromName(%s,%s) {index not found}",
	     domainName, mapIndexName);
      }
      else {
	log (LOG_DEBUG,
	     "dataDomainMapAndIndexFromName(%s,%s) {structures found}",
	     domainName, mapIndexName);
	dataMapLock (map);
	rc = 0;
      }
    }
  }
  *domainReturn = domain;
  *mapReturn = map;
  *indexReturn = index;
  return rc;
}
