/*
  heap.c
  Jeff McElroy
  jmcelroy11@sio.midco.net
*/

/*
    ypanything, an NIS gateway to non-NIS sources
    Copyright (C) 2004  Jeff McElroy <jmcelroy11@sio.midco.net>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                                                                 
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>

#include "log.h"
#include "heap.h"

static long G_allocCount = 0;

typedef struct HEAPNODE
{
  struct HEAPNODE *m_next;
}
HEAPNODE;

struct MEM_TYPE
{
  const char *m_name;
  const int m_keepHeap;
  int m_count;
  HEAPNODE *m_firstFree;
  pthread_mutex_t m_mutext;
} G_memType[MEMTYPE_COUNT] =
{
  "UNDEFINED", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "DATA", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "DOMAIN", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "DOMAINNAME", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "MAP", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "MAPNAME", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "INDEX", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "INDEXNAME", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "MAPPING", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "MAPPINGNODE", 1, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "CMDSOURCE_MAPTEXT", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "CMDSOURCE_MAPDATA", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "CMDSOURCE_SCRIPT", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "CMDSOURCE_DELIMINATOR", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "CMDSOURCE_THREADDATA", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "CMDSOURCE_INDEXDATA", 0, 0, NULL, PTHREAD_MUTEX_INITIALIZER,
    "RBNODE", 1, 0, NULL, PTHREAD_MUTEX_INITIALIZER};



static void *local_heapAlloc (size_t size, int type)
{
  HEAPNODE *heap;

  pthread_mutex_lock (&G_memType[type].m_mutext);
  if (G_memType[type].m_firstFree != NULL) {
    /* just recycle an old heap */
    heap = G_memType[type].m_firstFree;
    G_memType[type].m_firstFree = heap->m_next;
    pthread_mutex_unlock (&G_memType[type].m_mutext);
  }
  else {
    pthread_mutex_unlock (&G_memType[type].m_mutext);
    heap = (HEAPNODE *) malloc (size + sizeof (HEAPNODE));
    heap->m_next = NULL;
  }

  return heap + 1;
}

static void local_heapFree (void *p, int type)
{
  HEAPNODE *heap = p;
  heap--;

  pthread_mutex_lock (&G_memType[type].m_mutext);
  heap->m_next = G_memType[type].m_firstFree;
  G_memType[type].m_firstFree = heap;
  pthread_mutex_unlock (&G_memType[type].m_mutext);
}

void heapInitialize ()
{
  int i;
  for (i = 0; i < MEMTYPE_COUNT; i++) {
    G_memType[i].m_count = 0;
    G_memType[i].m_firstFree = NULL;
  }
  G_allocCount = 0;
}

void heapDump ()
{
  int i;
  for (i = 0; i < MEMTYPE_COUNT; i++) {
    log (LOG_MEMORY, "%s: %d", G_memType[i].m_name, G_memType[i].m_count);
  }
}

void *heapAlloc (size_t size, int type)
{
  void *ptr;

  if (++G_allocCount % 100000 == 0) {
    heapDump ();
  }

  if (type >= MEMTYPE_COUNT || type < 0) {
    log (LOG_CRIT, "illegal heap type %d", type);
  }

  G_memType[type].m_count++;

  if (G_memType[type].m_keepHeap) {
    ptr = local_heapAlloc (size, type);
  }
  else {
    ptr = malloc (size);
  }

  if (ptr == NULL) {
    log (LOG_CRIT, "out of memory for heap type %s", G_memType[type].m_name);
  }

  return ptr;
}

void heapFree (void *p, int type)
{
  if (type >= MEMTYPE_COUNT || type < 0) {
    log (LOG_CRIT, "illegal heap type %d", type);
  }
  G_memType[type].m_count--;

  if (G_memType[type].m_keepHeap) {
    local_heapFree (p, type);
  }
  else {
    free (p);
  }

}

void *heapRealloc (void *old, int size, int type)
{
  void *ptr;

  if (type >= MEMTYPE_COUNT || type < 0) {
    log (LOG_CRIT, "illegal heap type %d", type);
  }

  if (G_memType[type].m_keepHeap) {
    log (LOG_CRIT, "heapRealloc not supported for this type of data");
  }
  else {
    ptr = realloc (old, size);
  }

  if (ptr == NULL) {
    log (LOG_CRIT, "out of memory");
  }

  return ptr;
}

char *heapStrdup (char *psz, int type)
{
  void *ptr;

  if (type >= MEMTYPE_COUNT || type < 0) {
    log (LOG_CRIT, "illegal heap type %d", type);
  }

  G_memType[type].m_count++;
  ptr = strdup (psz);
  if (ptr == NULL) {
    log (LOG_CRIT, "heap out of memory");
  }

  return ptr;
}
