/*
  mapping.c
  Jeff McElroy
  jmcelroy11@sio.midco.net
*/

/*
    ypanything, an NIS gateway to non-NIS sources
    Copyright (C) 2004  Jeff McElroy <jmcelroy11@sio.midco.net>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                                                                 
*/

#include <stdlib.h>
#include <assert.h>

#include "heap.h"
#include "mapping.h"
#include "rb.h"

typedef int (*MAPPING_COMPARE_FUNC) (char *p1, char *p2);

struct MAPPING
{
  MAPPING_COMPARE_FUNC m_compare;
  Rb_node m_rbnRoot;
};

typedef struct NODE
{
  void *m_key;
  int m_keyLength;
  void *m_value;
  int m_valueLength;
} NODE;


static int compare (NODE * n1, NODE * n2)
{
  if (n1->m_keyLength < n2->m_keyLength) {
    return -1;
  }
  else if (n1->m_keyLength > n2->m_keyLength) {
    return 1;
  }
  else {
    return memcmp (n1->m_key, n2->m_key, n1->m_keyLength);
  }
}


MAPPING *mappingOpen ()
{
  MAPPING *mapping;

  mapping = (MAPPING *) heapAlloc (sizeof (MAPPING), MEMTYPE_MAPPING);
  assert (mapping != NULL);

  mapping->m_compare = (MAPPING_COMPARE_FUNC) compare;
  mapping->m_rbnRoot = make_rb ();

  return mapping;

}

void mappingClose (MAPPING * mapping)
{
  Rb_node rbnIndex;
  NODE *node;
  for
    (rbnIndex = rb_first (mapping->m_rbnRoot);
     rbnIndex != rb_nil (mapping->m_rbnRoot); rbnIndex = rb_next (rbnIndex)) {
    node = (NODE *) rbnIndex->k.key;
    /*
       heapFree(node->m_key, MEMTYPE_UNDEFINED);
       heapFree(node->m_value, MEMTYPE_UNDEFINED);
     */
    heapFree (node, MEMTYPE_MAPPINGNODE);
  }
  rb_free_tree (mapping->m_rbnRoot);
  heapFree (mapping, MEMTYPE_MAPPING);
}

int
mappingSet (MAPPING * mapping, void *key, int keyLength, void *value,
	    int valueLength)
{
  NODE *node;
  node = heapAlloc (sizeof (NODE), MEMTYPE_MAPPINGNODE);
  assert (node != NULL);

  node->m_key = key;
  node->m_keyLength = keyLength;
  node->m_value = value;
  node->m_valueLength = valueLength;

  rb_insertg (mapping->m_rbnRoot, (char *) node, NULL, mapping->m_compare);
  return 0;
}

int
mappingGet (MAPPING * mapping, void **key, int *keyLength, void **value,
	    int *valueLength)
{
  Rb_node rbnIndex;
  NODE node;
  NODE *nodeResult;


  if (*key == NULL) {
    return mappingGetFirst (mapping, key, keyLength, value, valueLength);
  }
  else {
    node.m_key = *key;
    node.m_keyLength = *keyLength;

    rbnIndex =
      rb_find_gkey (mapping->m_rbnRoot, (char *) &node, mapping->m_compare);
    if (rbnIndex != NULL
	&& mapping->m_compare (rbnIndex->k.key, (char *) &node) == 0) {
      nodeResult = (NODE *) rbnIndex->k.key;
      *value = nodeResult->m_value;
      *valueLength = nodeResult->m_valueLength;
      return 0;
    }
  }
  return 1;
}

int
mappingGetFirst (MAPPING * mapping, void **key, int *keyLength, void **value,
		 int *valueLength)
{
  Rb_node rbnIndex;
  Rb_node rbnEnd;
  NODE *nodeResult;

  rbnIndex = rb_first (mapping->m_rbnRoot);
  rbnEnd = rb_prev (rbnIndex);
  if (rbnIndex != NULL && rbnIndex != rbnEnd) {
    nodeResult = (NODE *) rbnIndex->k.key;
    *key = nodeResult->m_key;
    *keyLength = nodeResult->m_keyLength;
    *value = nodeResult->m_value;
    *valueLength = nodeResult->m_valueLength;
    return 0;
  }
  return 1;
}

int
mappingGetNext (MAPPING * mapping, void **key, int *keyLength, void **value,
		int *valueLength)
{
  Rb_node rbnIndex;
  NODE node;
  NODE *nodeResult;

  node.m_key = *key;
  node.m_keyLength = *keyLength;

  rbnIndex =
    rb_find_gkey (mapping->m_rbnRoot, (char *) &node, mapping->m_compare);
  if (rbnIndex != NULL
      && mapping->m_compare (rbnIndex->k.key, (char *) &node) == 0) {
    rbnIndex = rb_next (rbnIndex);
    if (rbnIndex != rb_nil (mapping->m_rbnRoot)) {
      nodeResult = (NODE *) rbnIndex->k.key;
      *key = nodeResult->m_key;
      *keyLength = nodeResult->m_keyLength;
      *value = nodeResult->m_value;
      *valueLength = nodeResult->m_valueLength;
      return 0;
    }
  }
  return 1;
}
